cpdef create_point_geometries(xarray, yarray)
cdef _create_point_geometries(xarray, yarray)
cpdef create_way_geometries(node_coordinates, way_elements)
cdef create_pygeos_polygon_from_relation(node_coordinates, relation_ways, member_roles)
cpdef create_polygon_geometries(node_coordinates, way_elements)
cdef create_linear_ring(coordinates)
cpdef create_node_coordinates_lookup(nodes)
cdef pygeos_to_shapely(geom)
cdef to_shapely(pygeos_array)
cdef get_way_coordinates_for_polygon(node_coordinate_lookup, way_records)
